from dataclasses import dataclass
from typing import Type, Dict, List, Union, TypeVar, Generic
import json


@dataclass
class NludbRequest():
  pass

@dataclass
class NludbResponseData():
  @staticmethod
  def safely_from_dict(d: any) -> Dict:
    """Last resort if subclass doesn't override: pass through."""
    return d

@dataclass
class TaskStatusResponse:
  taskId: str = None
  taskStatus: str = None
  taskCreatedOn: str = None
  taskLastModifiedOn: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "TaskStatusResponse":
    return TaskStatusResponse(
      taskId = d.get('taskId', None),
      taskStatus = d.get('taskStatus', None),
      taskCreatedOn = d.get('taskCreatedOn', None),
      taskLastModifiedOn = d.get('taskLastModifiedOn', None)
    )


T = TypeVar('T')      # Declare type variable

@dataclass
class NludbResponse(Generic[T]):
  task: TaskStatusResponse
  data: T

  def update(self, response: "AsyncTask"):
    if self.task is not None:
      return self.task.update(response)

  def wait(self, max_timeout_s: float=60, retry_delay_s: float=1):
    if self.task is not None:
      return self.task.wait(max_timeout_s=max_timeout_s, retry_delay_s=retry_delay_s)

  def check(self):
    if self.task is not None:
      return self.task.check()

  def add_comment(self, externalId: str = None, externalType: str = None, externalGroup: str = None, metadata: any = None) -> "NludbResponse[TaskCommentResponse]":
    if self.task is not None:
      return self.task.add_comment(externalId = externalId, externalType = externalType, externalGroup = externalGroup, metadata = metadata)

  def list_comments(self) -> "NludbResponse[ListTaskCommentResponse]":
    if self.task is not None:
      return self.task.list_comments()

  def delete_comment(self, taskCommentId: str = None) -> "NludbResponse[TaskCommentResponse]":
    if self.task is not None:
      return self.task.delete_comment(taskCommentId=taskCommentId)

Metadata = Union[int, float, bool, str, List, Dict]

def str_to_metadata(s: str) -> Metadata:
  if s is None:
    return None
  try:
    return json.loads(s)
  except:
    s

def metadata_to_str(m: Metadata) -> str:
  if m is None:
    return None
  try:
    return json.dumps(m)
  except:
    m



# from nludb.types.async_task import NludbTaskStatus
# import pytest
# from os import path
# from .helpers import _random_name, _nludb
# from nludb import NLUDB, BlockTypes, FileFormats

# __author__ = "Edward Benson"
# __copyright__ = "Edward Benson"
# __license__ = "MIT"


# def test_file_upload():
#     nludb = _nludb()
#     name_a = "{}.mkd".format(_random_name())
#     a = nludb.upload_file(
#         name=name_a,
#         content="A",
#         format=FileFormats.MKD
#     )
#     assert(a.id is not None)
#     assert(a.name == name_a)
#     assert(a.format == FileFormats.MKD)

#     name_b = "{}.txt".format(_random_name())
#     b = nludb.upload_file(
#         name=name_b,
#         content="B",
#         format=FileFormats.TXT
#     )
#     assert(b.id is not None)
#     assert(b.name == name_b)
#     assert(b.format == FileFormats.TXT)
#     assert(a.id != b.id)

#     name_c = "{}.txt".format(_random_name())
#     c = nludb.upload_file(
#         name=name_c,
#         content="B",
#         format=FileFormats.MKD
#     )
#     assert(c.format == FileFormats.MKD) # The specified format gets precedence over filename

#     d = nludb.upload_file(
#         name=name_c,
#         content="B",
#     )
#     assert(d.format == FileFormats.TXT) # The filename is used in a pinch.

#     a.delete()
#     b.delete()
#     c.delete()
#     d.delete()


# # def test_image_upload():
# #     nludb = _nludb()
# #     test_filename = path.join(
# #         path.dirname(path.realpath(__file__)),
# #         'test_img.png'
# #     )
# #     with open(test_filename, "rb") as f:
# #         name_c = "{}.png".format(_random_name())
# #         c = nludb.upload_file(
# #             name=name_c,
# #             content=f
# #         )
# #         assert(c.id is not None)
# #         assert(c.name == name_c)
# #         assert(c.format == FileFormats.PNG)
# #         c.delete()


# def test_file_scrape():
#     nludb = _nludb()

#     name_a = "{}.html".format(_random_name())
#     a = nludb.scrape_file(
#         name=name_a,
#         url="https://edwardbenson.com/2020/10/gpt3-travel-agent"
#     )
#     assert(a.id is not None)
#     assert(a.name == name_a)
#     assert(a.format == FileFormats.HTML)

#     name_b = "{}.html".format(_random_name())
#     b = nludb.scrape_file(
#         name=name_b,
#         url="https://edwardbenson.com/2018/09/case-of-the-murderous-ai"
#     )
#     assert(b.id is not None)    
#     assert(a.id != b.id)
#     assert(b.name == name_b)
#     assert(b.format == FileFormats.HTML)

#     a.delete()
#     b.delete()

# # def test_file_add_bloc():
# #     nludb = _nludb()

# #     name_a = "{}.txt".format(_random_name())
# #     a = nludb.upload_file(
# #         name=name_a,
# #         content="This is a test."
# #     )
# #     assert(a.id is not None)
# #     task  = a.convert()
# #     task._run_development_mode()
# #     task.wait()
# #     q1 = a.query()
# #     assert(len(q1.blocks) == 2)

# #     # TODO: Append Content
# #     # TODO: Append Blocks

# def test_file_upload_then_parse():
#     nludb = _nludb()

#     name_a = "{}.txt".format(_random_name())
#     a = nludb.upload_file(
#         name=name_a,
#         content="This is a test."
#     )
#     assert(a.id is not None)

#     q1 = a.query()
#     assert(len(q1.blocks) == 0)

#     task  = a.convert()
#     task.wait()

#     q1 = a.query()
#     assert(len(q1.blocks) == 2)
#     assert(q1.blocks[0].type == BlockTypes.Document)    
#     assert(q1.blocks[1].type == BlockTypes.Paragraph)    
#     assert(q1.blocks[1].value == 'This is a test.')

#     name_b = "{}.mkd".format(_random_name())
#     b = nludb.upload_file(
#         name=name_b,
#         content="""# Header

# This is a test."""
#     )
#     assert(b.id is not None)

#     q1 = b.query()
#     assert(len(q1.blocks) == 0)

#     task  = b.convert()
#     task._run_development_mode()
#     task.wait()

#     q1 = b.query()
#     assert(len(q1.blocks) == 3)
#     assert(q1.blocks[0].type == BlockTypes.Document)    
#     assert(q1.blocks[2].type == BlockTypes.Paragraph)
#     assert(q1.blocks[2].value == 'This is a test.')
#     assert(q1.blocks[1].type == BlockTypes.H1)    
#     assert(q1.blocks[1].value == 'Header')

#     q2 = b.query(blockType=BlockTypes.H1)
#     assert(len(q2.blocks) == 1)
#     assert(q2.blocks[0].type == BlockTypes.H1)
#     assert(q2.blocks[0].value == 'Header')

#     a.delete()
#     b.delete()



class ParsingModels:
  EN_DEFAULT = "sp_en_core_web_trf"
  SPACY_EN_CORE_WEB_TRF = "sp_en_core_web_trf"

class ParsingResultGranularity:
  SENTENCE = "sentence"
  TOKEN = "token"
  FULL_PARSE = "full_parse"



class BlockTypes:
  Document = "doc"
  Page = "page" # E.g. in a PDF
  Region = "region" # E.g., abstract catchall region in a document
  Header = "header"
  H1 = "h1"
  H2 = "h2"
  H3 = "h3"
  H4 = "h4"
  H5 = "h5"
  Line = "line"
  Title = "title"
  Subtitle = "subtitle"
  Footer = "footer"
  Paragraph = "paragraph"
  List = "list"
  Blockquote = "blockquote"
  Blockcode = "blockcode"
  Unk = "unk"
  Sentence = "sentence"
  Token = "token"
  Span = "span"



from typing import ItemsView
from nludb.types.async_task import NludbTaskStatus
from nludb.types.embedding_index import IndexItem
import pytest

from .helpers import _random_index, _random_name, _nludb, qa_model, sim_model

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def _list_equal(actual, expected):
    assert len(actual) == len(expected)
    assert all([a == b for a, b in zip(actual, expected)])

def test_basic_task_comment():
    nludb = _nludb()
    name = _random_name()
    with _random_index(nludb) as index:
        item1 = IndexItem(
          value="Pizza",
          externalId="pizza",
          externalType="food",
          metadata=[1,2,3]
        )

        index.insert(item1.value, externalId=item1.externalId, externalType=item1.externalType, metadata=item1.metadata)
        task = index.embed()
        task.wait()

        res2 = index.search(item1.value, includeMetadata=True, k=1)
        res2.add_comment(externalId = "Foo", externalType = "Bar", metadata=[1,2])
        # We don't return to Res2 until the end to make sure we aren't co-mingling comments!

        res = index.search(item1.value, includeMetadata=True, k=1)

        assert (res.data.hits is not None)
        assert (len(res.data.hits) == 1)
        assert (res.data.hits[0].value == item1.value)
        assert (res.data.hits[0].externalId == item1.externalId)
        assert (res.data.hits[0].externalType == item1.externalType)
        _list_equal(res.data.hits[0].metadata, item1.metadata)

        res.add_comment(externalId = "Foo", externalType = "Bar", metadata=[1,2])

        comments = res.list_comments()
        assert (len(comments.data.comments) == 1)

        comment = comments.data.comments[0]
        assert (comment.externalId == "Foo")
        assert (comment.externalType == "Bar")
        _list_equal(comment.metadata, [1, 2])

        res.delete_comment(comment.taskCommentId)

        comments = res.list_comments()
        assert (len(comments.data.comments) == 0)

        # Now let's add one
        res.add_comment(externalId = "Foo1", externalType = "Bar1", metadata=[1,2, 3])
        res.add_comment(externalId = "Foo2", externalType = "Bar2", metadata=[1,2, 3, 4])

        comments = res.list_comments()
        assert (len(comments.data.comments) == 2)

        comment = comments.data.comments[0]
        assert (comment.externalId == "Foo1")
        assert (comment.externalType == "Bar1")
        _list_equal(comment.metadata, [1, 2, 3])

        comment = comments.data.comments[1]
        assert (comment.externalId == "Foo2")
        assert (comment.externalType == "Bar2")
        _list_equal(comment.metadata, [1, 2, 3, 4])

        res.delete_comment(comments.data.comments[0].taskCommentId)
        res.delete_comment(comments.data.comments[1].taskCommentId)

        comments = res.list_comments()
        assert (len(comments.data.comments) == 0)

        # Now we handle res2
        comments = res2.list_comments()
        assert (len(comments.data.comments) == 1)
        comment = comments.data.comments[0]
        assert (comment.externalId == "Foo")
        assert (comment.externalType == "Bar")
        _list_equal(comment.metadata, [1, 2])
        res.delete_comment(comments.data.comments[0].taskCommentId)
        comments = res.list_comments()
        assert (len(comments.data.comments) == 0)


def test_task_comment_feedback_reporting():
  """
  We want to be able to generate reports like this:

  Select Across Gorup    -- externalGroup
  Inputs Seen: XXX       -- Distinct externalId
  Inputs Suggested: YYY  -- Add to metadata
  Inputs Liked / Disliked / Used -- Add to metadata

  So really we just need to test the group aggregation
  """
  nludb = _nludb()
  name = _random_name()
  with _random_index(nludb) as index:
    item1 = IndexItem(
      value="Pizza",
      externalId="pizza",
      externalType="food",
      metadata=[1,2,3]
    )

    G1 = _random_name()
    G2 = _random_name()

    index.insert(item1.value, externalId=item1.externalId, externalType=item1.externalType, metadata=item1.metadata)
    task = index.embed()
    task.wait()

    res = index.search(item1.value, includeMetadata=True, k=1)
    res.add_comment(externalId = "Foo1", externalType = "Bar1", externalGroup = G1, metadata=[1,2, 3])
    res.add_comment(externalId = "Foo2", externalType = "Bar1", externalGroup = G1, metadata=[1,2, 3])
    res.add_comment(externalId = "Foo2", externalType = "Bar1", externalGroup = G2, metadata=[1,2, 3])

    comments = res.list_comments()
    assert (len(comments.data.comments) == 3)

    g1 = nludb.tasks.list_comments(externalGroup=G1)
    assert (len(g1.data.comments) == 2)

    g2 = nludb.tasks.list_comments(externalGroup=G2)
    assert (len(g2.data.comments) == 1)

    g1 = nludb.tasks.list_comments(taskId=res.task.taskId, externalGroup=G1)
    assert (len(g1.data.comments) == 2)

    g2 = nludb.tasks.list_comments(taskId=res.task.taskId, externalGroup=G2)
    assert (len(g2.data.comments) == 1)

    g1 = nludb.tasks.list_comments(taskId=res.task.taskId, externalId="Foo1", externalGroup=G1)
    assert (len(g1.data.comments) == 1)

    g2 = nludb.tasks.list_comments(taskId=res.task.taskId, externalId="Foo1", externalGroup=G2)
    assert (len(g2.data.comments) == 0)

    res.delete_comment(comments.data.comments[0].taskCommentId)
    res.delete_comment(comments.data.comments[1].taskCommentId)
    res.delete_comment(comments.data.comments[2].taskCommentId)

    g1 = nludb.tasks.list_comments(externalGroup=G1)
    assert (len(g1.data.comments) == 0)

    g2 = nludb.tasks.list_comments(externalGroup=G2)
    assert (len(g2.data.comments) == 0)



from dataclasses import dataclass
from nludb.types.base import NludbResponse, NludbResponseData, NludbRequest, TaskStatusResponse, str_to_metadata
import json
from typing import List

class NludbTaskStatus:
  waiting = "waiting"
  running = "running"
  succeeded = "succeeded"
  failed = "failed"

@dataclass
class TaskRunRequest(NludbRequest):
  taskId: str

@dataclass
class TaskStatusRequest(NludbRequest):
  taskId: str

@dataclass
class AddTaskCommentRequest(NludbRequest):
  taskId: str
  externalId: str = None
  externalType: str = None
  externalGroup: str = None
  metadata: str = None
  upsert: bool = None

@dataclass
class DeleteTaskCommentRequest(NludbRequest):
  taskCommentId: str

@dataclass
class TaskCommentResponse(NludbRequest):
  userId: str = None
  taskId: str = None
  taskCommentId: str = None
  externalId: str = None
  externalType: str = None
  externalGroup: str = None
  metadata: any = None
  createdAt: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "TaskCommentResponse":
    return TaskCommentResponse(
      userId = d.get('userId', None),
      taskId = d.get('taskId', None),
      taskCommentId = d.get('taskCommentId', None),
      externalId = d.get('externalId', None),
      externalType = d.get('externalType', None),
      externalGroup = d.get('externalGroup', None),
      metadata = str_to_metadata(d.get("metadata", None)),
      createdAt = d.get('createdAt', None)
    )

@dataclass
class ListTaskCommentRequest(NludbRequest):
  taskId: str = None
  externalId: str = None
  externalType: str = None
  externalGroup: str = None

@dataclass
class ListTaskCommentResponse(NludbRequest):
  comments: List[TaskCommentResponse]

  @staticmethod
  def safely_from_dict(d: any) -> "ListTaskCommentResponse":
    return ListTaskCommentResponse(
      comments = [TaskCommentResponse.safely_from_dict(dd) for dd in d.get('comments', [])]
    )



import requests
import logging
import time
import os

from nludb import __version__
from nludb.types.base import NludbRequest, NludbResponseData, TaskStatusResponse, metadata_to_str
from dataclasses import asdict
from typing import Type, TypeVar, Generic
from nludb.types.async_task import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

import typing

T = typing.TypeVar('T', bound=NludbResponseData)

class AsyncTask(Generic[T]):
  """Encapsulates a unit of asynchronously performed work."""
  nludb: "ApiBase" = None
  taskId: str = None
  taskStatus: str = None
  taskCreatedOn: str = None
  taskLastModifiedOn: str = None

  def __init__(
    self, 
    nludb: "ApiBase" = None, 
    taskId: str = None, 
    taskStatus: str = None, 
    taskCreatedOn: str = None, 
    taskLastModifiedOn: str = None,
    eventualResultType: Type[NludbResponseData] = None
    ):
    self.nludb =  nludb
    self.taskId = taskId
    self.taskStatus = taskStatus
    self.taskCreatedOn = taskCreatedOn
    self.taskLastModifiedOn = taskLastModifiedOn
    self.eventualResultType = eventualResultType

  def update(self, response: TaskStatusResponse):
    """Incorporates a `TaskStatusResponse` into this object."""
    if response is not None:
      self.taskId = response.taskId
      self.taskStatus = response.taskStatus
      self.taskCreatedOn = response.taskCreatedOn
      self.taskLastModifiedOn = response.taskLastModifiedOn
    else:
      self.taskStatus = None
      
  def check(self):
    """Retrieves and incorporates a fresh status update."""
    req = TaskStatusRequest(
      self.taskId
    )
    resp = self.nludb.post(
      'task/status',
      payload=req,
      expect=TaskStatusResponse,
      asynchronous=True
    )
    self.update(resp.task)

  def add_comment(self, externalId: str = None, externalType: str = None, externalGroup: str = None, metadata: any = None, upsert: bool = True) -> NludbResponse[TaskCommentResponse]:
    req = AddTaskCommentRequest(
      taskId=self.taskId,
      externalId=externalId,
      externalType=externalType,
      externalGroup=externalGroup,
      metadata=metadata_to_str(metadata),
      upsert=upsert
    )
    return self.nludb.post(
      'task/comment/create',
      req,
      expect=TaskCommentResponse,
    )

  def list_comments(self) -> NludbResponse[ListTaskCommentResponse]:
    req = ListTaskCommentRequest(
      taskId=self.taskId,
    )
    return self.nludb.post(
      'task/comment/list',
      req,
      expect=ListTaskCommentResponse,
    )

  def delete_comment(self, taskCommentId: str = None) -> NludbResponse[TaskCommentResponse]:
    req = DeleteTaskCommentRequest(
      taskCommentId=taskCommentId
    )
    return self.nludb.post(
      'task/comment/delete',
      req,
      expect=TaskCommentResponse,
    )

  def wait(self, max_timeout_s: float=60, retry_delay_s: float=1):
    """Polls and blocks until the task has succeeded or failed (or timeout reached)."""
    start = time.time()
    self.check()
    if self.taskStatus == NludbTaskStatus.succeeded or self.taskStatus == NludbTaskStatus.failed:
      return
    time.sleep(retry_delay_s)

    while time.time() - start < max_timeout_s:
      time.sleep(retry_delay_s)
      self.check()
      if self.taskStatus == NludbTaskStatus.succeeded or self.taskStatus == NludbTaskStatus.failed:
        return
    
class ApiBase:
  """Base class for API connectivity. 
  
  Separated primarily as a hack to prevent ciruclar imports.
  """
  def __init__(
    self, 
    api_key: str=None, 
    api_domain: str=None,
    api_version: int=None):

    self.api_key = api_key
    if self.api_key is None:
      print("NONE")
      if 'NLUDB_KEY' in os.environ:
        self.api_key = os.getenv('NLUDB_KEY')
        print("SET")

    self.api_domain = api_domain
    if self.api_domain is None:
      if 'NLUDB_DOMAIN' in os.environ:
        self.api_domain = os.getenv('NLUDB_DOMAIN')
      else:
        self.api_domain = "https://api.nludb.com/"

    self.api_version = api_version
    if self.api_version is None:
      if 'NLUDB_VERSION' in os.environ:
        self.api_version = os.getenv('NLUDB_VERSION')
      else:
        self.api_version = 1

    separator = '/'
    if self.api_domain.endswith('/'):
      separator = ''
    self.endpoint = "{}{}api/v{}".format(
      self.api_domain, 
      separator,
      self.api_version
    )
  
  T = TypeVar('T', bound=NludbResponseData)

  def post(
    self, 
    operation: str, 
    payload: NludbRequest = None,
    file: None = None,
    expect: T = NludbResponseData,
    asynchronous: bool = False,
    debug: bool = False
  ) -> NludbResponse[T]:
    """Post to the NLUDB API.

    All responses have the format:
       {
         data: <actual response>,
         error?: {
           reason: message
         }
       }
    
    For the Python client we return the contents of the `data`
    field if present, and we raise an exception if the `error`
    field is filled in.
    """
    if self.api_key is None:
      raise Exception("Please set your NLUDB API key.")

    url = "{}/{}".format(self.endpoint, operation)
    if file is not None:
      data = asdict(payload) if payload is not None else {}

      # Note: requests seems to have a bug passing boolean (and maybe numeric?)
      # values in the midst of multipart form data. You need to manually convert
      # it to a string; otherwise it will pass as False or True (with the capital),
      # which is not standard notation outside of Python.
      for key in data:
        if data[key] is False:
          data[key] = 'false'
        elif data[key] is True:
          data[key] = 'true'
      resp = requests.post(
        url,
        files={"file": file},
        data=data,
        headers = {"Authorization": "Bearer {}".format(self.api_key)}
      )
    else:
      resp = requests.post(
        url,
        json=asdict(payload) if payload is not None else None,
        headers = {"Authorization": "Bearer {}".format(self.api_key)}
      )
    if debug is True:
      print("Response", resp)
    j = resp.json()
    if debug is True:
      print("Response JSON", j)
    
    # Error response
    if 'reason' in j:
      import json
      data = asdict(payload) if payload is not None else {}
      raise Exception(j['reason'])

    if 'data' not in j and 'status' not in j:
      raise Exception('No data or status property in response')

    task = None
    if 'status' in j:
      task_resp = TaskStatusResponse.safely_from_dict(j['status'])
      if task_resp is not None and task_resp.taskId is not None:
          task = AsyncTask(nludb=self)
          task.update(task_resp)

    obj = None
    if 'data' in j:
      obj = expect.safely_from_dict(j['data'])

    return NludbResponse[T](
      task=task,
      data=obj
    )



from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponseData
from nludb.types.parsing import DependencyMatcher, PhraseMatcher, TokenMatcher

class FileUploadType:
  file = "file"
  url = "url"

@dataclass
class FileUploadRequest(NludbRequest):
  type: str
  name: str
  fileFormat: str = None
  convert: bool = False

@dataclass
class FileUploadResponse(NludbResponseData):
  fileId: str
  fileFormat: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileUploadResponse":
    return FileUploadResponse(
      fileId = d.get('fileId', None),
      fileFormat = d.get('fileFormat', None)
    )

@dataclass
class FileDeleteRequest(NludbRequest):
  fileId: str

@dataclass
class FileDeleteResponse(NludbResponseData):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileDeleteResponse":
    return FileDeleteResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class FileConvertRequest(NludbRequest):
  fileId: str

@dataclass
class FileConvertResponse(NludbResponseData):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileConvertResponse":
    return FileConvertResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class FileParseRequest(NludbRequest):
  fileId: str
  model: str = None
  tokenMatchers: List[TokenMatcher] = None
  phraseMatchers: List[PhraseMatcher] = None
  dependencyMatchers: List[DependencyMatcher] = None

@dataclass
class FileParseResponse(NludbResponseData):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileParseResponse":
    return FileParseResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class Block(NludbRequest):
  blockId: str
  type: str
  value: str

  @staticmethod
  def safely_from_dict(d: any) -> "Block":
    return Block(
      blockId = d.get('blockId', None),
      type = d.get('type', None),
      value = d.get('value', None)
    )

@dataclass
class FileQueryRequest(NludbRequest):
  fileId: str
  blockType: str = None

@dataclass
class FileQueryResponse(NludbResponseData):
  fileId: str
  blocks: List[Block]

  @staticmethod
  def safely_from_dict(d: any) -> "FileQueryResponse":
    return FileQueryResponse(
      fileId = d.get('fileId', None),
      blocks = [Block.safely_from_dict(block) for block in d.get('blocks', None)]
    )



class FileFormats:
  UNKNOWN = "unknown"
  TXT = "txt"
  MKD = "mkd"
  EPUB = "epub"
  PDF = "pdf"
  JPG = "jpg"
  PNG = "png"
  TIFF = "tiff"
  GIF = "gif"
  HTML = "html"
  DOC = "doc"
  DOCX = "docx"
  PPT = "ppt"
  PPTX = "pptx"
  RTF = "rtf"



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .nludb import NLUDB
from .embedding_index import EmbeddingIndex
from .classifier import Classifier
from .types.embedding_models import EmbeddingModels
from .types.classifier_models import ClassifierModels
from .types.parsing_models import ParsingModels
from .file import File
from .types.model import ModelType, ModelAdapterType, LimitUnit, Model
from .types.block_types import BlockTypes
from .types.file_formats import FileFormats


from nludb.types.async_task import NludbTaskStatus
from nludb.types.embedding_index import IndexSnapshotRequest
import pytest
import os
import random
import string
import contextlib
import time

from nludb import NLUDB, EmbeddingModels, EmbeddingIndex
from .helpers import _random_index, _random_name, _nludb, qa_model

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def _insert(index, items):
  for item in items:
    index.insert(item, "TestId", "TestType", [1,2,3])

  # Now embed
  task = index.embed()
  task.wait()
  task.check()
  assert (task.task.taskStatus == NludbTaskStatus.succeeded)

def _snapshot(index, windowSize=None):
  if windowSize is None:
    task = index.create_snapshot()
    task.wait()
    task.check()
  else:
    # We do this manually so as not to clutter the end-user visible
    # API with debug/testing parameters
    req = IndexSnapshotRequest(
      index.id,
      windowSize=windowSize
    )
    task = index.nludb.post(
      'embedding-index/snapshot/create',
      req,
      expect=IndexSnapshotRequest,
      asynchronous=True
    )
    task.wait()
    task.check()

def test_snapshot_create():
  nludb = _nludb()

  name = _random_name()
  index = nludb.create_index(
      name=name,
      model=qa_model(),
      upsert=True
  )

  _insert(index, ["Oranges are orange."])
  search_results = index.search("What color are oranges?", includeMetadata=True)
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "index")
  assert(search_results.data.hits[0].value == "Oranges are orange.")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  _snapshot(index)
  search_results = index.search("What color are oranges?", includeMetadata=True)
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "snapshot")
  assert(search_results.data.hits[0].value == "Oranges are orange.")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  _insert(index, ["Apples are red."])
  search_results = index.search("What color are apples?", includeMetadata=True)
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "index")
  assert(search_results.data.hits[0].value == "Apples are red.")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  _snapshot(index)
  search_results = index.search("What color are apples?", includeMetadata=True)
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "snapshot")
  assert(search_results.data.hits[0].value == "Apples are red.")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  index.delete()

def test_snapshot_window():
  nludb = _nludb()

  name = _random_name()
  index = nludb.create_index(
      name=name,
      model=qa_model(),
      upsert=True
  )

  sentences = []
  for i in range(15):
    sentences.append("Orange number {} is as good as the last".format(i))

  SENT = "Is orange number 13 any good?"  
  _insert(index, sentences)

  search_results = index.search(SENT, includeMetadata=True).data
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "index")
  assert(search_results.data.hits[0].value == "Orange number 13 is as good as the last")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  _snapshot(index, windowSize=2)
  search_results = index.search(SENT, includeMetadata=True).data
  assert(len(search_results.data.hits) == 1)
  assert(search_results.data.hits[0].indexSource == "snapshot")
  assert(search_results.data.hits[0].value == "Orange number 13 is as good as the last")
  assert(search_results.data.hits[0].externalId == "TestId")
  assert(search_results.data.hits[0].externalType == "TestType")
  assert(len(search_results.data.hits[0].metadata)  == 3)

  index.delete()



import logging
import json
from typing import Union, List, Dict

from nludb import __version__
from nludb.api.base import ApiBase, AsyncTask, NludbResponse
from nludb.types.async_task import ListTaskCommentRequest, ListTaskCommentResponse
from nludb.types.model import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class Tasks:
  """Asynchronous background task (and task feedback) management.
  """

  def __init__(self, nludb: ApiBase):
    self.nludb = nludb

  def list_comments(
    self,
    taskId: str = None,
    externalId: str = None,
    externalType: str = None,
    externalGroup: str = None
  ) -> NludbResponse[ListTaskCommentResponse]:
    return self.nludb.post(
      'task/comment/list',
      ListTaskCommentRequest(
        taskId=taskId,
        externalId=externalId,
        externalType=externalType,
        externalGroup=externalGroup
      ),
      expect=ListTaskCommentResponse
    )



from typing import List, Dict, Callable
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponseData

@dataclass
class Token:
  text: str = None
  textWithWs: str = None
  whitespace: str = None
  head: str = None
  headI: int = None
  leftEdge: str = None
  rightEdge: str = None
  entType: str = None
  entIob: str = None
  lemma: str = None
  normalized: str = None
  shape: str = None
  prefix: str = None
  suffix: str = None
  isAlpha: bool = None
  isAscii: bool = None
  isDigit: bool = None
  isTitle: bool = None
  isPunct: bool = None
  isLeftPunct: bool = None
  isRightPunct: bool = None
  isSpace: bool = None
  isBracket: bool = None
  isQuote: bool = None
  isCurrency: bool = None
  likeUrl: bool = None
  likeNum: bool = None
  likeEmail: bool = None
  isOov: bool = None
  isStop: bool = None
  pos: str = None
  tag: str = None
  dep: str = None
  lang: str = None
  prob: float = None
  charIndex: int = None
  tokenIndex: int = None

  @staticmethod
  def safely_from_dict(d: any) -> "Token":
    return Token(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      whitespace = d.get("whitespace", None),
      head = d.get("head", None),
      headI = d.get("headI", None),
      leftEdge = d.get("leftEdge", None),
      rightEdge = d.get("rightEdge", None),
      entType = d.get("entType", None),
      entIob = d.get("entIob", None),
      lemma = d.get("lemma", None),
      normalized = d.get("normalized", None),
      shape = d.get("shape", None),
      prefix = d.get("prefix", None),
      suffix = d.get("suffix", None),
      isAlpha = d.get("isAlpha", None),
      isAscii = d.get("isAscii", None),
      isDigit = d.get("isDigit", None),
      isTitle = d.get("isTitle", None),
      isPunct = d.get("isPunct", None),
      isLeftPunct = d.get("isLeftPunct", None),
      isRightPunct = d.get("isRightPunct", None),
      isSpace = d.get("isSpace", None),
      isBracket = d.get("isBracket", None),
      isQuote = d.get("isQuote", None),
      isCurrency = d.get("isCurrency", None),
      likeUrl = d.get("likeUrl", None),
      likeNum = d.get("likeNum", None),
      likeEmail = d.get("likeEmail", None),
      isOov = d.get("isOov", None),
      isStop = d.get("isStop", None),
      pos = d.get("pos", None),
      tag = d.get("tag", None),
      dep = d.get("dep", None),
      lang = d.get("lang", None),
      prob = d.get("prob", None),
      charIndex = d.get("charIndex", None),
      tokenIndex = d.get("tokenIndex", None)
    )

  @staticmethod
  def from_spacy(t: any, includeParseData: bool=True) -> "Token":
    if includeParseData is False:
      return Token(
        text = t.text,
        textWithWs = t.text_with_ws,
        whitespace = t.whitespace_,
        leftEdge = t.left_edge.text,
        rightEdge = t.right_edge.text,
        lemma = t.lemma_,
        isOov = t.is_oov,
        isStop = t.is_stop,
        charIndex = t.idx,
        tokenIndex = t.i,
      )

    return Token(
      text = t.text,
      textWithWs = t.text_with_ws,
      whitespace = t.whitespace_,
      head = t.head.text,
      headI = t.head.i,
      leftEdge = t.left_edge.text,
      rightEdge = t.right_edge.text,
      entType = t.ent_type_,
      entIob = t.ent_iob_,
      lemma = t.lemma_,
      normalized = t.norm_,
      shape = t.shape_,
      prefix = t.prefix_,
      suffix = t.suffix_,
      isAlpha = t.is_alpha,
      isAscii = t.is_ascii,
      isDigit = t.is_digit,
      isTitle = t.is_title,
      isPunct = t.is_punct,
      isLeftPunct = t.is_left_punct,
      isRightPunct = t.is_right_punct,
      isSpace = t.is_space,
      isBracket = t.is_bracket,
      isQuote = t.is_quote,
      isCurrency = t.is_currency,
      likeUrl = t.like_url,
      likeNum = t.like_num,
      likeEmail = t.like_email,
      isOov = t.is_oov,
      isStop = t.is_stop,
      pos = t.pos_,
      tag = t.tag_,
      dep = t.dep_.upper(),
      lang = t.lang_,
      prob = t.prob,
      charIndex = t.idx,
      tokenIndex = t.i,
    )

@dataclass
class Entity:
  text: str
  textWithWs: str
  startChar: int
  endChar: int
  startToken: int
  endToken: int
  label: str
  lemma: str

  @staticmethod
  def safely_from_dict(d: any) -> "Entity":
    return Entity(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      startChar = d.get("startChar", None),
      endChar = d.get("endChar", None),
      startToken = d.get("startToken", None),
      endToken = d.get("endToken", None),
      label = d.get("label", None),
      lemma = d.get("lemma", None)
    )

  @staticmethod
  def from_spacy(ent: any) -> "Entity":
    return Entity(
      text=ent.text,
      textWithWs=ent.text_with_ws,
      startChar=ent.start_char,
      endChar=ent.end_char,
      startToken=ent.start,
      endToken=ent.end,
      label=ent.label_,
      lemma=ent.lemma_,
    )

@dataclass
class Span:
  text: str
  textWithWs: str
  startChar: int
  endChar: int
  startToken: int
  endToken: int
  label: str
  lemma: str
  sentiment: float

  @staticmethod
  def safely_from_dict(d: any) -> "Span":
    return Span(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      startChar = d.get("startChar", None),
      endChar = d.get("endChar", None),
      startToken = d.get("startToken", None),
      endToken = d.get("endToken", None),
      label = d.get("label", None),
      lemma = d.get("lemma", None),
      sentiment = d.get("sentiment", None)
    )

  @staticmethod
  def from_spacy(ent: any) -> "Span":
    return Span(
      text=ent.text,
      textWithWs=ent.text_with_ws,
      startChar=ent.start_char,
      endChar=ent.end_char,
      startToken=ent.start,
      endToken=ent.end,
      label=ent.label_,
      lemma=ent.lemma_,
      sentiment=ent.sentiment
    )

@dataclass
class Sentence:
  text: str
  tokens: List[Token]
  entities: List[Entity]
  spans: List[Span]
  sentiment: float

  @staticmethod
  def safely_from_dict(d: any) -> "Sentence":
    tokens = [Token.safely_from_dict(h) for h in (d.get("tokens", []) or [])]
    entities = [Entity.safely_from_dict(h) for h in (d.get("entities", []) or [])]
    spans = [Span.safely_from_dict(h) for h in (d.get("spans", []) or [])]

    return Sentence(
      text=d.get("text", None),
      tokens=tokens,
      entities=entities,
      spans=spans,
      sentiment = d.get("sentiment", None)
    )

  @staticmethod
  def from_spacy(d: any, includeTokens: bool=True, includeParseData: bool=True, includeEntities: bool=True) -> "Sentence":
    tokens = [Token.from_spacy(t, includeParseData=includeParseData) for t in d] if includeTokens is True else []
    entities = [Entity.from_spacy(e) for e in d.ents] if includeEntities is True else []

    return Sentence(
      text=d.text,
      tokens=tokens,
      entities=entities,
      spans=[],
      sentiment=d.sentiment
    )

@dataclass
class Doc:
  text: str
  sentences: List[Sentence]
  spans: List[Span]
  model: str
  lang: str
  sentiment: float
  entities: List[Entity]

  @staticmethod
  def safely_from_dict(d: any) -> "Doc":
    sentences = [Sentence.safely_from_dict(h) for h in (d.get("sentences", []) or [])]
    spans = [Span.safely_from_dict(h) for h in (d.get("spans", []) or [])]
    entities = [Entity.safely_from_dict(h) for h in (d.get("entities", []) or [])]

    return Doc(
      text=d.get("text", None),
      sentences=sentences,
      spans=spans,
      model=d.get("model", None),
      lang=d.get("lang", None),
      sentiment = d.get("sentiment", None),
      entities = entities
    )

  @staticmethod
  def from_spacy(
    text: str, 
    model: str, 
    d: any, 
    includeTokens: bool=True, 
    includeParseData: bool=True, 
    includeEntities: bool=True, 
    sentenceFilterFn: Callable[[any], bool]=None
    ) -> "Doc":
    sentences = [
      Sentence.from_spacy(s, includeTokens=includeTokens, includeParseData=includeParseData, includeEntities=includeEntities) 
      for s in d.sents
      if sentenceFilterFn is None or sentenceFilterFn(s) is True
    ]
    spans = []
    for label in d.spans:
      span_group = d.spans[label]
      for s in span_group:
        spans.append(Span.safely_from_dict(s))

    entities = []
    if includeEntities is True:
      for ent in d.ents:
        e = Entity.from_spacy(ent)
        if e is not None:
          entities.append(e)

    ret = Doc(
      text=text,
      sentences=sentences,
      spans=spans,
      model=model,
      lang=d.lang_,
      sentiment=d.sentiment,
      entities=entities
    )
    return ret
  
@dataclass
class ParseResponse(NludbResponseData):
  docs: List[Doc] = None

  @staticmethod
  def safely_from_dict(d: any) -> "ParseResponse":
    docs = [Doc.safely_from_dict(h) for h in (d.get("docs", []) or [])]
    return ParseResponse(
      docs=docs
    )

MatcherClause = Dict[str, any]
Matcher = List[MatcherClause]

@dataclass
class TokenMatcher():
  label: str = None
  patterns: List[Matcher] = None

  @staticmethod
  def safely_from_dict(d: any) -> "TokenMatcher":
    return TokenMatcher(
      label=d.get("label", None),
      patterns=(d.get("patterns", []) or [])
    )

@dataclass
class PhraseMatcher():
  label: str = None
  phrases: List[str] = None
  attr: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "PhraseMatcher":
    return PhraseMatcher(
      label=d.get("label", None),
      phrases=(d.get("phrases", []) or []),
      attr=d.get("attr", None)
    )

@dataclass
class DependencyMatcher():
  label: str = None
  # Note: Add a LABEL field to the matcher clause to have a token
  # get labeled in the response.
  patterns: List[Matcher] = None

  @staticmethod
  def safely_from_dict(d: any) -> "DependencyMatcher":
    return DependencyMatcher(
      label=d.get("label", None),
      patterns=(d.get("patterns", []) or [])
    )

@dataclass
class ParseRequest(NludbResponseData):
  docs: List[str] = None
  model: str = None
  tokenMatchers: List[TokenMatcher] = None
  phraseMatchers: List[PhraseMatcher] = None
  dependencyMatchers: List[DependencyMatcher] = None

  includeTokens: bool = True
  includeParseData: bool = True
  includeEntities: bool = False

  metadata: any = None

  @staticmethod
  def safely_from_dict(d: any) -> "ParseRequest":
    token_matchers = [TokenMatcher.safely_from_dict(h) for h in (d.get("tokenMatchers", []) or [])]
    phrase_matchers = [PhraseMatcher.safely_from_dict(h) for h in (d.get("phraseMatchers", []) or [])]
    dependency_matchers = [DependencyMatcher.safely_from_dict(h) for h in (d.get("dependencyMatchers", []) or [])]

    includeTokens = d.get("includeTokens", True)
    if includeTokens is None:
      includeTokens = True

    includeParseData = d.get("includeParseData", True)
    if includeParseData is None:
      includeParseData = True

    includeEntities = d.get("includeEntities", True)
    if includeEntities is None:
      includeEntities = True

    metadata = d.get("metadata", None)
    if metadata is not None:
      try:
        metadata = json.loads(metadata)
      except:
        pass

    return ParseRequest(
      docs=(d.get("docs", []) or []),
      model=d.get("model", None),
      tokenMatchers=token_matchers,
      phraseMatchers=phrase_matchers,
      dependencyMatchers=dependency_matchers,
      includeTokens=includeTokens,
      includeParseData=includeParseData,
      includeEntities=includeEntities,
      metadata=metadata
    )




# from nludb.types.async_task import NludbTaskStatus
# from nludb.types.parsing_models import ParsingModels
# import pytest
# from os import path
# from .helpers import _random_name, _nludb
# from nludb import NLUDB, BlockTypes, FileFormats

# __author__ = "Edward Benson"
# __copyright__ = "Edward Benson"
# __license__ = "MIT"

# # TODO: It should fail if the docs field is empty.
# # TODO: It should fail if the file hasn't been converted.

# def test_file_parse():
#   nludb = _nludb()
#   name_a = "{}.mkd".format(_random_name())
#   T = "This is the title"
#   P1_1 = "This is the first sentence of paragraph 1."
#   P1_2 = "This is the second sentence of paragraph 1."
#   P2_1 = "This is the first sentence of paragraph 2."
#   P2_2 = "This is the second sentence of paragraph 2."

#   a = nludb.upload_file(
#     name=name_a,
#     content="# {}\n\n{} {}\n\n{} {}".format(T, P1_1, P1_2, P2_1, P2_2),
#     format=FileFormats.MKD
#   )
#   assert(a.id is not None)
#   assert(a.name == name_a)
#   assert(a.format == FileFormats.MKD)

#   task  = a.convert()
#   task.wait()

#   q1 = a.query(blockType=BlockTypes.H1)
#   assert(len(q1.blocks) == 1)
#   assert(q1.blocks[0].type == BlockTypes.H1)
#   assert(q1.blocks[0].value == T)

#   q2 = a.query(blockType=BlockTypes.Paragraph)
#   assert(len(q2.blocks) == 2)
#   assert(q2.blocks[0].type == BlockTypes.Paragraph)
#   assert(q2.blocks[0].value == "{} {}".format(P1_1, P1_2))

#   # The sentences aren't yet parsed out!
#   q2 = a.query(blockType=BlockTypes.Sentence)
#   assert(len(q2.blocks) == 0)

#   # Now we parse
#   task = a.parse(model=ParsingModels.EN_DEFAULT)
#   task.wait()

#   # Now the sentences should be parsed!
#   q2 = a.query(blockType=BlockTypes.Sentence)
#   assert(len(q2.blocks) == 5) # The 5th is inside the header!

#   a.delete()

    


# from nludb.types.async_task import NludbTaskStatus
# from nludb.types.parsing_models import ParsingModels
# import pytest
# from os import path
# from .helpers import _random_name, _nludb
# from nludb import NLUDB, BlockTypes, FileFormats
# from .helpers import _random_index, _random_name, _nludb

# __author__ = "Edward Benson"
# __copyright__ = "Edward Benson"
# __license__ = "MIT"

# # TODO: It should fail if the docs field is empty.
# # TODO: It should fail if the file hasn't been converted.

# def test_file_parse():
#   nludb = _nludb()
#   name_a = "{}.mkd".format(_random_name())
#   T = "A poem about roses"
#   P1_1 = "Roses are red."
#   P1_2 = "Violets are blue."
#   P2_1 = "Sugar is sweet."
#   P2_2 = "I love you."
#   T2 = "A story about cake"
#   P3_1 = "Cake is made of flour."
#   P3_2 = "Cake tastes good with milk."
#   P4_1 = "Cake comes in chocolate and vanilla flavors."
#   P4_2 = "Cake can be cut into many pieces and shared."
  
#   content1 = "# {}\n\n{} {}\n\n{} {}".format(T, P1_1, P1_2, P2_1, P2_2)
#   content2 = "# {}\n\n{} {}\n\n{} {}".format(T2, P3_1, P3_2, P4_1, P4_2)
#   content = "{}\n\n{}".format(content1, content2)

#   a = nludb.upload_file(
#     name=name_a,
#     content=content,
#     format=FileFormats.MKD
#   )
#   assert(a.id is not None)
#   assert(a.name == name_a)
#   assert(a.format == FileFormats.MKD)

#   task  = a.convert()
#   task.wait()

#   # Now we parse
#   task = a.parse(model=ParsingModels.EN_DEFAULT)
#   task.wait()

#   # Now the sentences should be parsed!
#   q2 = a.query(blockType=BlockTypes.Sentence)
#   assert(len(q2.blocks) == 10) # The 5th is inside the header!

#   # Now we add the file to the index
#   with _random_index(nludb) as index:
#     insert_results = index.insert_file(a.id, reindex=False)
#     task = index.embed()
#     task.wait()

#     res = index.search("What color are roses?")
#     assert(len(res.hits) == 1)
#     assert(res.hits[0].value == P1_1)

#   a.delete()

  


from nludb.types.async_task import NludbTaskStatus
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, ParsingModels, EmbeddingModels, EmbeddingIndex, ClassifierModels, File

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"


def _env_or(env_var: str, or_val: str) -> str:
    if env_var in os.environ:
        return os.getenv(env_var)
    return or_val

def _random_name() -> str:
    letters = string.digits + string.ascii_letters
    id =''.join(random.choice(letters) for i in range(10))
    return "test_{}".format(id)

def qa_model() -> str:
    return _env_or('NLUDB_EMBEDDER_QA', EmbeddingModels.QA)

def sim_model() -> str:
    return _env_or('NLUDB_EMBEDDER_SIM', EmbeddingModels.SIMILARITY)

def parsing_model() -> str:
    return _env_or('NLUDB_PARSER_DEFAULT', ParsingModels.EN_DEFAULT)

def zero_shot_model() -> str:
    return _env_or('NLUDB_CLASSIFIER_DEFAULT', ClassifierModels.HF_ZERO_SHOT_LBART)


@contextlib.contextmanager
def _random_index(nludb: NLUDB, model: str = qa_model()) -> EmbeddingIndex:
    index = nludb.create_index(
        name=_random_name(),
        model=model,
        upsert=True
    )
    yield index
    index.delete()  # or whatever you need to do at exit

@contextlib.contextmanager
def _random_file(nludb: NLUDB, content: str = "") -> File:
    file = nludb.create_file(
      name=_random_name(),
      contents=content
    )
    yield file
    file.delete()  # or whatever you need to do at exit

def _nludb() -> NLUDB:
    # This should automatically pick up variables from the environment.
    return NLUDB()


from typing import List
import json
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponseData, metadata_to_str
from nludb.types.search import Hit
@dataclass
class IndexCreateRequest(NludbRequest):
  name: str
  model: str
  upsert: bool = True
  externalId: str = None
  externalType: str = None
  metadata: any = None

@dataclass
class IndexCreateResponse(NludbResponseData):
  id: str

@dataclass
class IndexItem:
  value: str = None
  externalId: str = None
  externalType: str = None
  metadata: any = None

  def clone_for_insert(self) -> "IndexItem":
    """Produces a clone with a string representation of the metadata"""
    ret = IndexItem(
      value=self.value,
      externalId=self.externalId,
      externalType=self.externalType,
      metadata=self.metadata
    )
    if isinstance(ret.metadata, dict) or isinstance(ret.metadata, list):
      ret.metadata = json.dumps(ret.metadata)
    return ret

@dataclass
class IndexInsertRequest(NludbRequest):
  indexId: str
  items: List[IndexItem] = None
  value: str = None
  fileId: str = None
  blockType: str = None
  externalId: str = None
  externalType: str = None
  metadata: any = None
  reindex: bool = True

@dataclass
class IndexItemId(NludbResponseData):
  indexId: str = None
  id: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexItemId":
    return IndexItemId(
      indexId = d.get('indexId', None),
      id = d.get('id', None)
    )

@dataclass
class IndexInsertResponse(NludbResponseData):
  itemIds: List[IndexItemId] = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexInsertResponse":
    return IndexInsertResponse(
      itemIds = [IndexItemId.safely_from_dict(x) for x in (d.get('itemIds', []) or [])]
    )

@dataclass
class IndexEmbedRequest(NludbRequest):
  indexId: str

@dataclass
class IndexEmbedResponse(NludbResponseData):
  indexId: str

  @staticmethod
  def safely_from_dict(d: any) -> "IndexEmbedResponse":
    return IndexEmbedResponse(
      indexId = d.get('indexId', None)
    )

@dataclass
class IndexDeleteRequest(NludbRequest):
  indexId: str

@dataclass
class IndexDeleteResponse(NludbResponseData):
  indexId: str

  @staticmethod
  def safely_from_dict(d: any) -> "IndexDeleteResponse":
    return IndexDeleteResponse(
      indexId = d.get('indexId', None)
    )

@dataclass
class IndexSearchRequest(NludbRequest):
  indexId: str
  query: str = None
  queries: List[str] = None
  k: int = 1  
  includeMetadata: bool = False

@dataclass
class IndexSearchResponse(NludbResponseData):
  hits: List[Hit] = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexSearchResponse":
    hits = [Hit.safely_from_dict(h) for h in (d.get("hits", []) or [])]
    return IndexSearchResponse(
      hits=hits
    )

@dataclass
class IndexSnapshotRequest(NludbRequest):
  indexId: str
  # This variable is intended only to support
  # unit testing.
  windowSize: int = None

@dataclass
class IndexSnapshotResponse(NludbResponseData):
  indexId: str
  snapshotId: str

  @staticmethod
  def safely_from_dict(d: any) -> "IndexSnapshotResponse":
    return IndexSnapshotResponse(
      indexId = d.get('indexId', None),
      snapshotId = d.get('snapshotId', None)
    )

@dataclass
class ListSnapshotsRequest(NludbRequest):
  indexId: str = None

@dataclass
class ListSnapshotsResponse(NludbResponseData):
  snapshots: List[IndexSnapshotResponse]
  
  @staticmethod
  def safely_from_dict(d: any) -> "IndexSnapshotResponse":
    return IndexSnapshotResponse(
      snapshots = [IndexSnapshotResponse.safely_from_dict(dd) for dd in (d.get('snapshots', []) or [])]
    )

@dataclass
class DeleteSnapshotsRequest(NludbRequest):
  snapshotId: str = None

@dataclass
class DeleteSnapshotsResponse(NludbRequest):
  snapshotId: str = None
